from typing import List, Union
import torch
import pandas as pd
import numpy as np
from lightwood.encoder.base import BaseEncoder
from lightwood.api import dtype
from lightwood.encoder.helpers import MinMaxNormalizer, CatNormalizer
from lightwood.helpers.general import is_none


class ArrayEncoder(BaseEncoder):
    """
    Fits a normalizer for array data. To encode, `ArrayEncoder` returns a normalized window of previous data.
    It can be used for generic arrays, as well as for handling historical target values in time series tasks.

    Currently supported normalizing strategies are minmax for numerical arrays, and a simple one-hot for categorical arrays. See `lightwood.encoder.helpers` for more details on each approach.

    :param stop_after: time budget in seconds.
    :param window: expected length of array data.
    :param original_dtype: element-wise data type
    """  # noqa

    is_trainable_encoder: bool = True

    def __init__(self, stop_after: float, window: int = None, is_target: bool = False, original_type: dtype = None):
        super().__init__(is_target)
        self.stop_after = stop_after
        self.original_type = original_type
        self._normalizer = None
        if window is not None:
            self.output_size = window + 1
        else:
            self.output_size = None

    def _pad_and_strip(self, array: List[object]):
        if len(array) < self.output_size:
            array = array + [0] * (self.output_size - len(array))
        if len(array) > self.output_size:
            array = array[:self.output_size]
        return array

    def prepare(self, train_priming_data, dev_priming_data):
        priming_data = pd.concat([train_priming_data, dev_priming_data])
        priming_data = priming_data.values

        if self.output_size is None:
            self.output_size = np.max([len(x) for x in priming_data if x is not None])
        for i in range(len(priming_data)):
            if is_none(priming_data[i]):
                priming_data[i] = [0] * self.output_size

        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        if self.original_type in (dtype.categorical, dtype.binary):
            self._normalizer = CatNormalizer(encoder_class='ordinal')
        else:
            self._normalizer = MinMaxNormalizer()

        if isinstance(priming_data, pd.Series):
            priming_data = priming_data.values

        priming_data = [self._pad_and_strip(list(x)) for x in priming_data]

        self._normalizer.prepare(priming_data)
        self.output_size *= self._normalizer.output_size
        self.is_prepared = True

    def encode(self, column_data: Union[list, np.ndarray, torch.Tensor]) -> torch.Tensor:
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if isinstance(column_data, pd.Series):
            column_data = column_data.values

        for i in range(len(column_data)):
            if is_none(column_data[i]):
                column_data[i] = [0] * self.output_size
        column_data = [self._pad_and_strip(list(x)) for x in column_data]

        data = torch.cat([self._normalizer.encode(column_data)], dim=-1)
        data[torch.isnan(data)] = 0.0
        data[torch.isinf(data)] = 0.0

        return data

    def decode(self, data) -> torch.tensor:
        decoded = data.tolist()
        return decoded



from sklearn.metrics import r2_score as sk_r2_score


def r2_score(y_true, y_pred) -> float:
    """ Wrapper for sklearn R2 score, lower capped between 0 and 1"""
    acc = sk_r2_score(y_true, y_pred)
    # Cap at 0
    if acc < 0:
        acc = 0
    # Guard against overflow (> 1 means overflow of negative score)
    if acc > 1:
        acc = 0

    return acc



import numpy as np
import pandas as pd
from typing import List
import torch
from lightwood.encoder.base import BaseEncoder


class IdentityEncoder(BaseEncoder):
    """
    The identity encoder which does not change the data during encoding.

    Due to the restrictions of torch.Tensor, the encoding function only accepts types:

    - (nested) lists of numbers
    - np.ndarray of numbers, or
    - torch.Tensor

    Nonetypes are automatically converted into nans. This cannot be decoded.
    If self.handle_nan is True, then all nans are converted to zeros and not decoded back.

    The decode function takes in a torch.Tensor and converts it to a list of numbers.
    """  # noqa

    def __init__(self, is_target=False, handle_nan=True) -> None:
        super().__init__(is_target)
        self.handle_nan = handle_nan

    # Not all encoders need to be prepared
    def prepare(self, priming_data: pd.Series) -> None:
        self._prepared = True

    def encode(self, column_data: object) -> torch.Tensor:
        if isinstance(column_data, torch.Tensor):
            res = column_data
        else:
            res = np.array(column_data, dtype=float)  # convert None to nan
            res = torch.Tensor(res)
        if self.handle_nan:
            res = torch.nan_to_num(res)
        return res

    def decode(self, encoded_data: torch.Tensor) -> List[object]:
        return encoded_data.tolist()



import torch
import numpy as np
from scipy.special import softmax
from lightwood.encoder.text.helpers.rnn_helpers import Lang
from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder

UNCOMMON_WORD = '__mdb_unknown_cat'
UNCOMMON_TOKEN = 0


class OneHotEncoder(BaseEncoder):
    """

    Why are we handling target weighting inside encoders? Simple: we'd otherwise have to compute per-index weighting inside the mixers, rather than having that code unified inside 2x encoders. So moving this to the mixer will still involve having to pass the target encoder to the mixer, but will add the additional complexity of having to pass a weighting map to the mixer and adding class-to-index translation boilerplate + weight setting for each mixer
    """ # noqa
    def __init__(self, is_target=False, target_weights=None, handle_unknown='unknown_token'):
        super().__init__(is_target)
        self._lang = None
        self.rev_map = {}

        if handle_unknown not in {"unknown_token", "return_zeros"}:
            raise ValueError(f"handle_unknown should be either 'unknown_token' or 'return_zeros', got {handle_unknown}")
        else:
            self.handle_unknown = handle_unknown

        if self.is_target:
            if self.handle_unknown != 'unknown_token':
                raise ValueError(f'One Hot Encoders used for target encoding can only be used with `handle_unknown` \
                                   set to `unknown_token`. The option: "{self.handle_unknown}" is not supported!')

        self.target_weights = None
        self.index_weights = None
        if self.is_target:
            self.target_weights = target_weights

    def prepare(self, priming_data, max_dimensions=20000):
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self._lang = Lang('default')
        if self.handle_unknown == "return_zeros":
            priming_data = [x for x in priming_data if x is not None]
            self._lang.index2word = {}
            self._lang.word2index = {}
            self._lang.n_words = 0
        elif self.handle_unknown == "unknown_token":
            priming_data = [x if x is not None else UNCOMMON_WORD for x in priming_data]
            self._lang.index2word = {UNCOMMON_TOKEN: UNCOMMON_WORD}
            self._lang.word2index = {UNCOMMON_WORD: UNCOMMON_TOKEN}
            self._lang.word2count[UNCOMMON_WORD] = 0
            self._lang.n_words = 1

        for category in priming_data:
            if category is not None:
                self._lang.addWord(str(category))

        while self._lang.n_words > max_dimensions:
            if self.handle_unknown == "return_zeros":
                necessary_words = []
            elif self.handle_unknown == "unknown_token":
                necessary_words = [UNCOMMON_WORD]
            least_occuring_words = self._lang.getLeastOccurring(n=len(necessary_words) + 1)

            word_to_remove = None
            for word in least_occuring_words:
                if word not in necessary_words:
                    word_to_remove = word
                    break

            self._lang.removeWord(word_to_remove)

        # Note: Is target assume that we are operating in "unknown_token" mode
        if self.is_target:
            self.index_weights = [1] * self._lang.n_words
            if self.target_weights is not None:
                uncommon_weight = np.min(list(self.target_weights.values()))
                self.index_weights[0] = uncommon_weight
                self.target_weights[UNCOMMON_WORD] = uncommon_weight
            for word in set(priming_data):
                if self.target_weights is not None:
                    self.index_weights[self._lang.word2index[str(word)]] = 1 / self.target_weights[word]

            self.index_weights = torch.Tensor(self.index_weights)

        self.output_size = self._lang.n_words
        self.rev_map = self._lang.index2word
        self.is_prepared = True

    def encode(self, column_data):
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        ret = []
        v_len = self._lang.n_words

        for word in column_data:
            encoded_word = [0] * v_len
            if word is not None:
                word = str(word)
                if self.handle_unknown == "return_zeros":
                    if word in self._lang.word2index:
                        index = self._lang.word2index[word]
                        encoded_word[index] = 1
                    else:
                        # Encoding an unknown value will result in a vector of zeros
                        log.warning('Trying to encode a value never seen before, returning vector of zeros')
                else:  # self.handle_unknown == "unknown_token"
                    index = self._lang.word2index[word] if word in self._lang.word2index else UNCOMMON_TOKEN
                    encoded_word[index] = 1

            ret.append(encoded_word)

        return torch.Tensor(ret)

    def decode(self, encoded_data, return_raw=False):
        encoded_data_list = encoded_data.tolist()
        ret = []
        probs = []

        for vector in encoded_data_list:
            # Logits and onehots are not the same in definition
            # But this explicitly operates on logits; it will take care of
            # the one hot (so you can pass something in the softmax logit space)
            # But will not affect something that is already OHE.

            all_zeros = not np.any(vector)
            if self.handle_unknown == "return_zeros" and all_zeros:
                ret.append(UNCOMMON_WORD)
            else:  # self.handle_unknown == "unknown_token"
                ohe_index = np.argmax(vector)
                ret.append(self._lang.index2word[ohe_index])

            if return_raw:
                probs.append(softmax(vector).tolist())

        if return_raw:
            return ret, probs, self.rev_map
        else:
            return ret



from typing import List
import torch


class BaseEncoder:
    """
    Base class for all encoders.
    
    An encoder should return encoded representations of any columnar data.
    The procedure for this is defined inside the `encode()` method.
    
    If this encoder is expected to handle an output column, then it also needs to implement the respective `decode()` method that handles the inverse transformation from encoded representations to the final prediction in the original column space.
    
    For encoders that learn representations (as opposed to rule-based), the `prepare()` method will handle all learning logic.
    
    The `to()` method is used to move PyTorch-based encoders to and from a GPU.
    
    :param  is_target: Whether the data to encode is the target, as per the problem definition.
    :param is_timeseries_encoder: Whether encoder represents sequential/time-series data. Lightwood must provide specific treatment for this kind of encoder
    :param is_trainable_encoder: Whether the encoder must return learned representations. Lightwood checks whether this flag is present in order to pass data to the feature representation via the ``prepare`` statement. 
    
    Class Attributes:
    - is_prepared: Internal flag to signal that the `prepare()` method has been successfully executed.
    - is_nn_encoder: Whether the encoder is neural network-based.
    - dependencies: list of additional columns that the encoder might need to encode.
    - output_size: length of each encoding tensor for a single data point.
    
    """ # noqa
    is_target: bool
    is_prepared: bool

    is_timeseries_encoder: bool = False
    is_trainable_encoder: bool = False

    def __init__(self, is_target=False) -> None:
        self.is_target = is_target
        self.is_prepared = False
        self.dependencies = []
        self.output_size = None

    # Not all encoders need to be prepared
    def prepare(self, priming_data) -> None:
        self.is_prepared = True

    def encode(self, column_data) -> torch.Tensor:
        raise NotImplementedError

    def decode(self, encoded_data) -> List[object]:
        raise NotImplementedError

    # Should work for all torch-based encoders, but custom behavior may have to be implemented for weird models
    def to(self, device, available_devices):
        # Find all nn.Module type objects and convert them
        # @TODO: Make this work recursively
        for v in vars(self):
            attr = getattr(self, v)
            if isinstance(attr, torch.nn.Module):
                attr.to(device)
        return self



import copy
import datetime
import dateutil
import numpy as np
import pandas as pd
import multiprocessing as mp
from lightwood.helpers.parallelism import get_nr_procs
from functools import partial
from typing import Dict
from lightwood.api.types import TimeseriesSettings
from lightwood.helpers.log import log
from lightwood.api import dtype


def transform_timeseries(
        data: pd.DataFrame, dtype_dict: Dict[str, str],
        timeseries_settings: TimeseriesSettings, target: str, mode: str) -> pd.DataFrame:
    """
    Block that transforms the dataframe of a time series task to a convenient format for use in posterior phases like model training.
    
    The main transformations performed by this block are:
      - Type casting (e.g. to numerical for `order_by` columns).
      - Windowing functions for historical context based on `TimeseriesSettings.window` parameter.
      - Explicitly add target columns according to the `TimeseriesSettings.nr_predictions` parameter.
      - Flag all rows that are "predictable" based on all `TimeseriesSettings`.
      - Plus, handle all logic for the streaming use case (where forecasts are only emitted for the last observed data point).
    
    :param data: Dataframe with data to transform.
    :param dtype_dict: Dictionary with the types of each column.
    :param timeseries_settings: A `TimeseriesSettings` object.
    :param target: The name of the target column to forecast.
    :param mode: Either "train" or "predict", depending on what phase is calling this procedure.
    
    :return: A dataframe with all the transformations applied.
    """  # noqa

    tss = timeseries_settings
    original_df = copy.deepcopy(data)
    gb_arr = tss.group_by if tss.group_by is not None else []
    ob_arr = tss.order_by
    window = tss.window

    if tss.use_previous_target and target not in data.columns:
        raise Exception(f"Cannot transform. Missing historical values for target column {target} (`use_previous_target` is set to True).")  # noqa

    for hcol in tss.historical_columns:
        if hcol not in data.columns or data[hcol].isna().any():
            raise Exception(f"Cannot transform. Missing values in historical column {hcol}.")

    if '__mdb_make_predictions' in original_df.columns:
        index = original_df[original_df['__mdb_make_predictions'].map(
            {'True': True, 'False': False, True: True, False: False}).isin([True])]
        infer_mode = index.shape[0] == 0  # condition to trigger: __mdb_make_predictions is set to False everywhere
        # @TODO: dont drop and use instead of original_index?
        original_df = original_df.reset_index(drop=True) if infer_mode else original_df
    else:
        infer_mode = False

    original_index_list = []
    idx = 0
    for row in original_df.itertuples():
        if _make_pred(row) or infer_mode:
            original_index_list.append(idx)
            idx += 1
        else:
            original_index_list.append(None)

    original_df['original_index'] = original_index_list

    secondary_type_dict = {}
    for col in ob_arr:
        if dtype_dict[col] in (dtype.date, dtype.integer, dtype.float):
            secondary_type_dict[col] = dtype_dict[col]

    # Convert order_by columns to numbers (note, rows are references to mutable rows in `original_df`)
    for _, row in original_df.iterrows():
        for col in ob_arr:
            # @TODO: Remove if the TS encoder can handle `None`
            if row[col] is None or pd.isna(row[col]):
                row[col] = 0.0
            else:
                if dtype_dict[col] == dtype.date:
                    try:
                        row[col] = dateutil.parser.parse(
                            row[col],
                            **{}
                        )
                    except (TypeError, ValueError):
                        pass

                if isinstance(row[col], datetime.datetime):
                    row[col] = row[col].timestamp()

                try:
                    row[col] = float(row[col])
                except ValueError:
                    raise ValueError(f'Failed to order based on column: "{col}" due to faulty value: {row[col]}')

    for oby in tss.order_by:
        original_df[f'__mdb_original_{oby}'] = original_df[oby]

    group_lengths = []
    if len(gb_arr) > 0:
        df_arr = []
        for _, df in original_df.groupby(gb_arr):
            df_arr.append(df.sort_values(by=ob_arr))
            group_lengths.append(len(df))
    else:
        df_arr = [original_df]
        group_lengths.append(len(original_df))

    n_groups = len(df_arr)
    last_index = original_df['original_index'].max()
    for i, subdf in enumerate(df_arr):
        if '__mdb_make_predictions' in subdf.columns and mode == 'predict':
            if infer_mode:
                df_arr[i] = _ts_infer_next_row(subdf, ob_arr, last_index)
                last_index += 1

    if len(original_df) > 500:
        # @TODO: restore possibility to override this with args
        nr_procs = get_nr_procs(original_df)
        log.info(f'Using {nr_procs} processes to reshape.')
        pool = mp.Pool(processes=nr_procs)
        # Make type `object` so that dataframe cells can be python lists
        df_arr = pool.map(partial(_ts_to_obj, historical_columns=ob_arr + tss.historical_columns), df_arr)
        df_arr = pool.map(partial(_ts_order_col_to_cell_lists,
                          order_cols=ob_arr + tss.historical_columns), df_arr)
        df_arr = pool.map(
            partial(
                _ts_add_previous_rows, order_cols=ob_arr + tss.historical_columns, window=window),
            df_arr)

        df_arr = pool.map(partial(_ts_add_future_target, target=target, nr_predictions=tss.nr_predictions,
                                  data_dtype=tss.target_type, mode=mode),
                          df_arr)

        if tss.use_previous_target:
            df_arr = pool.map(
                partial(_ts_add_previous_target, target=target, window=tss.window),
                df_arr)
        pool.close()
        pool.join()
    else:
        for i in range(n_groups):
            df_arr[i] = _ts_to_obj(df_arr[i], historical_columns=ob_arr + tss.historical_columns)
            df_arr[i] = _ts_order_col_to_cell_lists(df_arr[i], order_cols=ob_arr + tss.historical_columns)
            df_arr[i] = _ts_add_previous_rows(df_arr[i],
                                              order_cols=ob_arr + tss.historical_columns, window=window)
            df_arr[i] = _ts_add_future_target(df_arr[i], target=target, nr_predictions=tss.nr_predictions,
                                              data_dtype=tss.target_type, mode=mode)
            if tss.use_previous_target:
                df_arr[i] = _ts_add_previous_target(df_arr[i], target=target, window=tss.window)

    combined_df = pd.concat(df_arr)

    if '__mdb_make_predictions' in combined_df.columns:
        combined_df = pd.DataFrame(combined_df[combined_df['__mdb_make_predictions'].astype(bool).isin([True])])
        del combined_df['__mdb_make_predictions']

    if not infer_mode and any([i < tss.window for i in group_lengths]):
        if tss.allow_incomplete_history:
            log.warning("Forecasting with incomplete historical context, predictions might be subpar")
        else:
            raise Exception(f'Not enough historical context to make a timeseries prediction. Please provide a number of rows greater or equal to the window size. If you can\'t get enough rows, consider lowering your window size. If you want to force timeseries predictions lacking historical context please set the `allow_incomplete_history` timeseries setting to `True`, but this might lead to subpar predictions.') # noqa

    df_gb_map = None
    if n_groups > 1:
        df_gb_list = list(combined_df.groupby(tss.group_by))
        df_gb_map = {}
        for gb, df in df_gb_list:
            df_gb_map['_' + '_'.join(str(gb))] = df

    timeseries_row_mapping = {}
    idx = 0

    if df_gb_map is None:
        for _, row in combined_df.iterrows():
            if not infer_mode:
                timeseries_row_mapping[idx] = int(
                    row['original_index']) if row['original_index'] is not None and not np.isnan(
                    row['original_index']) else None
            else:
                timeseries_row_mapping[idx] = idx
            idx += 1
    else:
        for gb in df_gb_map:
            for _, row in df_gb_map[gb].iterrows():
                if not infer_mode:
                    timeseries_row_mapping[idx] = int(
                        row['original_index']) if row['original_index'] is not None and not np.isnan(
                        row['original_index']) else None
                else:
                    timeseries_row_mapping[idx] = idx

                idx += 1

    del combined_df['original_index']

    return combined_df


def _ts_infer_next_row(df: pd.DataFrame, ob: str, last_index: int) -> pd.DataFrame:
    """
    Adds an inferred next row for streaming mode purposes.

    :param df: dataframe from which next row is inferred.
    :param ob: `order_by` column.
    :param last_index: index number of the latest row in `df`.

    :return: Modified `df` with the inferred row appended to it.
    """
    last_row = df.iloc[[-1]].copy()
    if df.shape[0] > 1:
        butlast_row = df.iloc[[-2]]
        delta = (last_row[ob].values - butlast_row[ob].values).flatten()[0]
    else:
        delta = 1
    last_row.original_index = None
    last_row.index = [last_index + 1]
    last_row['__mdb_make_predictions'] = True
    last_row['__mdb_ts_inferred'] = True
    last_row[ob] += delta
    return df.append(last_row)


def _make_pred(row) -> bool:
    """
    Indicates whether a prediction should be made for `row` or not.
    """
    return not hasattr(row, '__mdb_make_predictions') or row.make_predictions


def _ts_to_obj(df: pd.DataFrame, historical_columns: list) -> pd.DataFrame:
    """
    Casts all historical columns in a dataframe to `object` type.

    :param df: Input dataframe
    :param historical_columns: Historical columns to type cast

    :return: Dataframe with `object`-typed historical columns
    """
    for hist_col in historical_columns:
        df.loc[:, hist_col] = df[hist_col].astype(object)
    return df


def _ts_order_col_to_cell_lists(df: pd.DataFrame, order_cols: list) -> pd.DataFrame:
    """
    Casts all data in `order_by` columns into cells.

    :param df: Input dataframe
    :param order_cols: `order_by` columns

    :return: Dataframe with all `order_cols` modified so that their values are cells, e.g. `1` -> `[1]`
    """
    for order_col in order_cols:
        for ii in range(len(df)):
            label = df.index.values[ii]
            df.at[label, order_col] = [df.at[label, order_col]]
    return df


def _ts_add_previous_rows(df: pd.DataFrame, order_cols: list, window: int) -> pd.DataFrame:
    """
    Adds previous rows (as determined by `TimeseriesSettings.window`) into the cells of all `order_by` columns.

    :param df: Input dataframe.
    :param order_cols: `order_by` columns.
    :param window: value of `TimeseriesSettings.window` parameter.
    
    :return: Dataframe with all `order_cols` modified so that their values are now arrays of historical context.
    """  # noqa
    for order_col in order_cols:
        for i in range(len(df)):
            previous_indexes = [*range(max(0, i - window), i)]

            for prev_i in reversed(previous_indexes):
                df.iloc[i][order_col].append(
                    df.iloc[prev_i][order_col][-1]
                )

            # Zero pad
            # @TODO: Remove since RNN encoder can do without (???)
            df.iloc[i][order_col].extend(
                [0] * (1 + window - len(df.iloc[i][order_col]))
            )
            df.iloc[i][order_col].reverse()
    return df


def _ts_add_previous_target(df: pd.DataFrame, target: str, window: int) -> pd.DataFrame:
    """
    Adds previous rows (as determined by `TimeseriesSettings.window`) into the cells of the target column.

    :param df: Input dataframe.
    :param target: target column name.
    :param window: value of `TimeseriesSettings.window` parameter.

    :return: Dataframe with new `__mdb_ts_previous_{target}` column that contains historical target context.
    """  # noqa
    if target not in df:
        return df
    previous_target_values = list(df[target])
    del previous_target_values[-1]
    previous_target_values = [None] + previous_target_values

    previous_target_values_arr = []
    for i in range(len(previous_target_values)):
        prev_vals = previous_target_values[max(i - window, 0):i + 1]
        arr = [None] * (window - len(prev_vals) + 1)
        arr.extend(prev_vals)
        previous_target_values_arr.append(arr)

    df[f'__mdb_ts_previous_{target}'] = previous_target_values_arr
    return df


def _ts_add_future_target(df, target, nr_predictions, data_dtype, mode):
    """
    Adds as many columns to the input dataframe as the forecasting horizon asks for (as determined by `TimeseriesSettings.nr_predictions`).

    :param df: Input dataframe.
    :param target: target column name.
    :param nr_predictions: value of `TimeseriesSettings.nr_predictions` parameter.
    :param data_dtype: dictionary with types of all input columns
    :param mode: either "train" or "predict". `Train` will drop rows with incomplet target info. `Predict` has no effect, for now.

    :return: Dataframe with new `{target}_timestep_{i}'` columns that contains target labels at timestep `i` of a total `TimeseriesSettings.nr_predictions`.
    """  # noqa
    if target not in df:
        return df
    if data_dtype in (dtype.integer, dtype.float, dtype.array, dtype.tsarray):
        df[target] = df[target].astype(float)

    for timestep_index in range(1, nr_predictions):
        next_target_value_arr = list(df[target])
        for del_index in range(0, min(timestep_index, len(next_target_value_arr))):
            del next_target_value_arr[0]
            next_target_value_arr.append(None)
        col_name = f'{target}_timestep_{timestep_index}'
        df[col_name] = next_target_value_arr
        df[col_name] = df[col_name].fillna(value=np.nan)

    # drop rows with incomplete target info.
    if mode == 'train':
        for col in [f'{target}_timestep_{i}' for i in range(1, nr_predictions)]:
            if '__mdb_make_predictions' not in df.columns:
                df['__mdb_make_predictions'] = True
            df.loc[df[col].isna(), ['__mdb_make_predictions']] = False

    return df



import torch
import numpy as np
from scipy.special import softmax
from lightwood.encoder.base import BaseEncoder


# Exists mainly for datasets with loads of binary flags where OHE can be too slow to fit
class BinaryEncoder(BaseEncoder):
    """

    Why are we handling target weighting inside encoders? Simple: we'd otherwise have to compute per-index weighting inside the mixers, rather than having that code unified inside 2x encoders. So moving this to the mixer will still involve having to pass the target encoder to the mixer, but will add the additional complexity of having to pass a weighting map to the mixer and adding class-to-index translation boilerplate + weight setting for each mixer
    """ # noqa
    def __init__(self, is_target=False, target_weights=None):
        super().__init__(is_target)
        self.map = {}
        self.rev_map = {}
        self.output_size = 2

        self.target_weights = None
        self.index_weights = None
        if self.is_target:
            self.target_weights = target_weights

    def prepare(self, priming_data):
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        for x in priming_data:
            x = str(x)
            if x not in self.map:
                self.map[x] = len(self.map)
                self.rev_map[len(self.rev_map)] = x

            if len(self.map) == 2:
                break

        if self.is_target:
            self.index_weights = [None, None]
            for word in self.map:
                if self.target_weights is not None:
                    self.index_weights[self.map[word]] = 1 / self.target_weights[word]
                else:
                    self.index_weights[self.map[word]] = 1

            self.index_weights = torch.Tensor(self.index_weights)

        self.is_prepared = True

    def encode(self, column_data):
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')
        ret = []

        for word in column_data:
            index = self.map.get(word, None)
            ret.append([0, 0])
            if index is not None:
                ret[-1][index] = 1

        return torch.Tensor(ret)

    def decode(self, encoded_data, return_raw=False):
        encoded_data_list = encoded_data.tolist()
        ret = []
        probs = []

        for vector in encoded_data_list:
            ret.append(self.rev_map[np.argmax(vector)])

            if return_raw:
                probs.append(softmax(vector).tolist())

        if return_raw:
            return ret, probs, self.rev_map
        else:
            return ret



from typing import List
import logging
import torch
import torchvision.transforms as transforms
from PIL import Image
import pandas as pd
from lightwood.encoder.image.helpers.img_to_vec import Img2Vec
from lightwood.encoder.base import BaseEncoder


class Img2VecEncoder(BaseEncoder):
    """
    This encoder generates encoded representations for images using a pre-trained deep neural network.

    All input images are rescaled to a standard size of 224x224, and normalized using the mean and standard deviation of the ImageNet dataset (as it was used to train the underlying NN).
    
    Note that this encoder does not have a .decode() method yet. As such, models that predict images as output are not supported at this time. 
    
    For more information about the neural network this encoder uses, refer to the `lightwood.encoder.image.helpers.img_to_vec.Img2Vec`.
    """  # noqa

    is_trainable_encoder: bool = True

    def __init__(self, stop_after: float = 3600, is_target: bool = False):
        """
        :param stop_after: time budget, in seconds. 
        :param is_target: whether the encoder corresponds to the target column. This is not currently possible for Img2VecEncoder.
        """  # noqa
        assert not is_target
        super().__init__(is_target)
        self.is_prepared = False

        self._scaler = transforms.Resize((224, 224))
        self._normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        self._to_tensor = transforms.ToTensor()
        self._img_to_tensor = transforms.Compose([
            self._scaler,
            self._to_tensor,
            self._normalize
        ])
        self.stop_after = stop_after

        pil_logger = logging.getLogger('PIL')
        pil_logger.setLevel(logging.ERROR)

    def prepare(self, train_priming_data: pd.Series, dev_priming_data: pd.Series):
        # @TODO: finetune here? depending on time aim
        """
        Instances an `Img2Vec` object and sets the expected size for encoded representations.
        """
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self.model = Img2Vec()
        self.output_size = self.model.output_size
        self.is_prepared = True

    def to(self, device, available_devices):
        """
        Moves the model to-and-from CPU and GPU.

        :param device: will move the model to this device.
        :param available_devices: all available devices as reported by lightwood.

        :return: same object but moved to the target device.
        """
        self.model.to(device, available_devices)
        return self

    def encode(self, images: List[str]) -> torch.Tensor:
        """
        Creates encodings for a list of images; each image is referenced by a filepath or url.

        :param images: list of images, each image is a path to a file or a url.
        :return: a torch.floatTensor
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        img_tensors = [self._img_to_tensor(
            Image.open(img_path)
        ) for img_path in images]
        vec_arr = []

        self.model.eval()
        with torch.no_grad():
            for img_tensor in img_tensors:
                vec = self.model(img_tensor.unsqueeze(0), batch=False)
                vec_arr.append(vec)
        return torch.stack(vec_arr).to('cpu')

    def decode(self, encoded_values_tensor):
        raise Exception('This encoder is not bi-directional')
